#ifndef COGL_TEST_DECLARATIONS_H
#define COGL_TEST_DECLARATIONS_H

void test_pipeline_user_matrix (void);
void test_blend_strings (void);
void test_blend (void);
void test_premult (void);
void test_path (void);
void test_path_clip (void);
void test_depth_test (void);
void test_backface_culling (void);
void test_layer_remove (void);
void test_sparse_pipeline (void);
void test_npot_texture (void);
void test_sub_texture (void);
void test_pixel_buffer_map (void);
void test_pixel_buffer_set_data (void);
void test_pixel_buffer_sub_region (void);
void test_wrap_modes (void);
void test_texture_get_set_data (void);
void test_atlas_migration (void);
void test_read_texture_formats (void);
void test_write_texture_formats (void);
void test_alpha_textures (void);
void test_primitive (void);
void test_just_vertex_shader (void);
void test_pipeline_uniforms (void);
void test_snippets (void);
void test_custom_attributes (void);
void test_offscreen (void);
void test_framebuffer_get_bits (void);
void test_point_size (void);
void test_point_size_attribute (void);
void test_point_size_attribute_snippet (void);
void test_point_sprite (void);
void test_point_sprite_orientation (void);
void test_point_sprite_glsl (void);
void test_alpha_test (void);
void test_map_buffer_range (void);
void test_primitive_and_journal (void);
void test_copy_replace_texture (void);
void test_pipeline_cache_unrefs_texture (void);
void test_pipeline_shader_state (void);
void test_gles2_context (void);
void test_gles2_context_fbo (void);
void test_gles2_context_copy_tex_image (void);
void test_euler (void);
void test_color_hsl (void);
void test_fence (void);
void test_texture_no_allocate (void);
void test_texture_rg (void);

#endif /* COGL_TEST_DECLARATIONS_H */
